{{ define "main" }}
<h1>Rclone Commands</h1>
<p>
  This is an index of all commands in rclone. Run <code>rclone
  command --help</code> to see the help for that command.
</p>
<table>
  <thead>
    <tr>
      <th>Command</th>
      <th>Description</th>
    </tr>
  </thead>
  <tbody>
    {{ range .Data.Pages }}
    <tr>
      <td><a href="{{ .RelPermalink }}">{{ .Title | markdownify }}</a></td>
      <td>{{ .Description  | markdownify }}</td>
    </tr>
    {{ end }}
  </tbody>
</table>
{{ end }}
