<li class="list-group-item d-flex justify-content-between py-1">
  <span>
    {{ .Get "name" }}{{ if .Get "note"}} (<a href="{{ .Get "config" }}{{ .Get "note" }}">See note</a>){{end}}
  </span>
  <span>
    <a class="badge badge-primary badge-pill" role="button btn-sm" href="{{ .Get "home" }}" target="_blank"><i class="fa fa-home" aria-hidden="true"></i> Home</a>
    <a class="badge badge-primary badge-pill" role="button btn-sm" href="{{ .Get "config" }}"><i class="fa fa-book" aria-hidden="true"></i> Config</a>
  </span>
</li>
