// Code generated by "stringer -type=vState"; DO NOT EDIT.

package vfs

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[vOK-0]
	_ = x[vAdd-1]
	_ = x[vDel-2]
}

const _vState_name = "vOKvAddvDel"

var _vState_index = [...]uint8{0, 3, 7, 11}

func (i vState) String() string {
	if i >= vState(len(_vState_index)-1) {
		return "vState(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _vState_name[_vState_index[i]:_vState_index[i+1]]
}
